/*
 * Copyright (C) 2014 Nicola De Maio
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package pomocov;

import beast.core.CalculationNode;
import beast.core.Description;
import beast.core.Input;
import beast.core.Input.Validate;
import beast.core.Loggable;
import beast.core.parameter.RealParameter;

import java.io.PrintStream;

import static pomocov.utils.IndicesEtc.*;
//import java.util.ArrayList;
//import java.util.List;

//import org.jblas.DoubleMatrix;
//import org.jblas.MatrixFunctions;


//import org.jblas.DoubleMatrix;
//import org.jblas.ComplexDoubleMatrix;
//import org.jblas.MatrixFunctions;
//import org.jblas.Eigen;





/**
 * @author Nicola De Maio 2015
 */
@Description("Describe the mutation model to be used (GTR, HKY, F81, general) within PoMo")
public class MutationModel extends CalculationNode implements Loggable {

    final static String []  pomocovModels = new String []  {"F81", "HKY", "GTR", "General", "StrandSymmetric", "StrandSymmetric_CTcontextfree", "StrandSymmetric_CTnbrsadd"};

    public Input<RealParameter> rateVectorInput = new Input<RealParameter>(
            "rateVector",
            "Mutation rate vector, any length among 1, 2, 6, 12 depending on the model that one wants to use. In case of 12, nucleotide frequencies are not used.",
            Validate.REQUIRED);
    public Input<RealParameter> nucFreqsInput = new Input<RealParameter>(
            "nucFreqs",
            "nucleotide frequencies as A, C, G, T.",
            Validate.REQUIRED);

    final public Input<String> modelTypeInput = new Input<>("modelType",
            "Mutation model type",
            "HKY",
            pomocovModels);

    public Input<RealParameter> otherRatesInput = new Input<>(
            "otherRates",
            "Rates for mutations in addition to strand-symmetric mutations");

    final public Input<Boolean> mutationOnlyModelInput = new Input<>(
            "mutationOnlyModel",
            "Use mutation-only within hosts (default false)");

    
    private RealParameter rateVector;
    private RealParameter nucFreqs;
    private RealParameter otherRates;

    public boolean mutationOnlyModel = false;


    enum MutationModelType {
        F81, HKY, GTR, General, StrandSymmetric, StrandSymmetric_CTcontextfree, StrandSymmetric_CTnbrsadd
    };
    private MutationModelType mmType;



    @Override
    public void initAndValidate() throws IllegalArgumentException {

        rateVector = rateVectorInput.get();
        nucFreqs = nucFreqsInput.get();

        rateVector.setLower(0.0);
        nucFreqs.setLower(0.0);
        int nCoV = 0;
        if (modelTypeInput.get() != null) {
            String sm = modelTypeInput.get();
            if (sm.compareTo("StrandSymmetric") == 0) {
                mmType = MutationModelType.StrandSymmetric;
                nCoV++;
            } else if (sm.compareTo("StrandSymmetric_CTcontextfree") == 0) {
                mmType = MutationModelType.StrandSymmetric_CTcontextfree;
                nCoV++;
            } else if (sm.compareTo("StrandSymmetric_CTnbrsadd") == 0) {
                mmType = MutationModelType.StrandSymmetric_CTnbrsadd;
                nCoV++;
            } else if (sm.compareTo("F81") == 0) {
                mmType = MutationModelType.F81;
                if (rateVector.getDimension() != 1) {
                    throw new IllegalArgumentException("MutationModelType.F81 requires rateVector to have dimension one");
                }
            } else if (sm.compareTo("HKY") == 0) {
                mmType = MutationModelType.HKY;
                if (rateVector.getDimension() != 2) {
                    throw new IllegalArgumentException("MutationModelType.HKY requires rateVector to have dimension 2");
                }
            } else if (sm.compareTo("GTR") == 0) {
                mmType = MutationModelType.GTR;
                if (rateVector.getDimension() != 6) {
                    throw new IllegalArgumentException("MutationModelType.GTR requires rateVector to have dimension 6");
                }
            } else if (sm.compareTo("General") == 0) {
                mmType = MutationModelType.General;
                if (rateVector.getDimension() != 12) {
                    throw new IllegalArgumentException("MutationModelType.General requires rateVector to have dimension 12");
                }
            } else {
                throw new IllegalArgumentException("modelType unknown");
            }
        } else {
            throw new IllegalArgumentException("The mutation model type must be supplied (eg modelType=\"HKY\").");
        }
        assert nCoV <= 1;
        if (mmType == MutationModelType.StrandSymmetric_CTcontextfree) {
            // GRJ TODO
            otherRates = otherRatesInput.get();
            if (otherRates == null) {
                throw  new IllegalArgumentException("SARSCoV2_CTcontextfree requires an input otherRates");
            }
            if (otherRates.getDimension() != 1) {
                throw  new IllegalArgumentException("otherRates must have 1 dimension for SARSCoV2_CTcontextfree ");
            }
        }
        mutationOnlyModel = false;
        if (mutationOnlyModelInput.get() != null) {
            mutationOnlyModel = mutationOnlyModelInput.get();
        }
        if (nucFreqs.getDimension() != 4) {
            throw new IllegalArgumentException("4 nucleotide frequencies must be provided.");
        }
    }


    /**
     * @return scaled mutation rates. Order: A->C, A->G, A->T, C->A, etc...
     */
    public double[] getACtoTGMutationRates() {
    	double[] rates = new double[12] ;
    	double[] values= rateVector.getDoubleValues() ;
        double[] freqs= getNucleotideFrequencies();
        
        switch (mmType) {
	        case F81:
	        	rates[0]=freqs[1]*values[0];
	        	rates[1]=freqs[2]*values[0];
	        	rates[2]=freqs[3]*values[0];
	        	rates[3]=freqs[0]*values[0];
	        	rates[4]=freqs[2]*values[0];
	        	rates[5]=freqs[3]*values[0];
	        	rates[6]=freqs[0]*values[0];
	        	rates[7]=freqs[1]*values[0];
	        	rates[8]=freqs[3]*values[0];
	        	rates[9]=freqs[0]*values[0];
	        	rates[10]=freqs[1]*values[0];
	        	rates[11]=freqs[2]*values[0];
	            break;
	        case HKY:
	        	rates[0]=freqs[1]*values[0];
	        	rates[1]=freqs[2]*values[1];
	        	rates[2]=freqs[3]*values[0];
	        	rates[3]=freqs[0]*values[0];
	        	rates[4]=freqs[2]*values[0];
	        	rates[5]=freqs[3]*values[1];
	        	rates[6]=freqs[0]*values[1];
	        	rates[7]=freqs[1]*values[0];
	        	rates[8]=freqs[3]*values[0];
	        	rates[9]=freqs[0]*values[0];
	        	rates[10]=freqs[1]*values[1];
	        	rates[11]=freqs[2]*values[0];
	            break;
	        case GTR:
	        	rates[0]=freqs[1]*values[0];
	        	rates[1]=freqs[2]*values[1];
	        	rates[2]=freqs[3]*values[2];
	        	rates[3]=freqs[0]*values[0];
	        	rates[4]=freqs[2]*values[3];
	        	rates[5]=freqs[3]*values[4];
	        	rates[6]=freqs[0]*values[1];
	        	rates[7]=freqs[1]*values[3];
	        	rates[8]=freqs[3]*values[5];
	        	rates[9]=freqs[0]*values[2];
	        	rates[10]=freqs[1]*values[4];
	        	rates[11]=freqs[2]*values[5];
	            break;
	        case General:
	        	for (int i=0;i<12;i++) rates[i]=values[i];
	            break;
            case StrandSymmetric:
                // just  strand-symmetric rates
                makeStrandSymmetricRates(rates, values);
                break;
            case StrandSymmetric_CTcontextfree:
                // strand-symmetric rates plus extra rate for C->T
                makeStrandSymmetricRates(rates, values);
                rates[rCT] += otherRates.getArrayValue(0);
                break;
            case StrandSymmetric_CTnbrsadd:
                assert false;
                makeStrandSymmetricRates(rates, values);
                // GRJ TODO. idea is an additive nbrs model with 8 parameters.
                break;
        }
        return rates;
    }


    private void makeStrandSymmetricRates(double[] rates, double[] values) {
        assert rates.length == 12;
        assert values.length == 6;
        for (int i = 0; i < 6; i++) {
            rates[i] = rates[11-i] = values[i];
        }
        assert  rates[rAC] == rates[rTG];
        assert  rates[rAG] == rates[rTC];
        assert  rates[rAT] == rates[rTA];
        assert  rates[rCA] == rates[rGT];
        assert  rates[rCG] == rates[rGC];
        assert  rates[rCT] == rates[rGA];
    }



    // GRJ The normalisation for linear approx for SARSCoV2 rates is completely different from BADTRIP/PoMo.
    // The mutation-only matrix is normalised as if it was acting on 4 nucleotides, like
    // a standard HKY/GTR type model.
    public double [] getACtoTGratesNormalisedForLinearApprox() {
        double [] rates = getACtoTGMutationRates();
        double[] freqs= getNucleotideFrequencies();
        double sum = 0.0;
        sum += freqs[nucleotideA] * (rates[rAC] + rates[rAG] + rates[rAT]);
        sum += freqs[nucleotideC] * (rates[rCA] + rates[rCG] + rates[rCT]);
        sum += freqs[nucleotideG] * (rates[rGA] + rates[rGC] + rates[rGT]);
        sum += freqs[nucleotideT] * (rates[rTA] + rates[rTC] + rates[rTG]);
        for (int i = 0; i < 12; i++) {
            rates[i] /= sum;
        }
        return rates;
    }


    public double[] getNucleotideFrequencies() {
    	double[] freqs=nucFreqs.getDoubleValues() ;
    	double sum=0.0;
    	for (int i=0; i<4; i++) sum+=freqs[i];
    	for (int i=0; i<4; i++) freqs[i]=freqs[i]/sum;
        return freqs ;
    }

    /*
     * Methods implementing loggable interface
     */
    
    @Override
    public void init(PrintStream out) throws IllegalArgumentException {
        String outName;
        if (getID() == null || getID().matches("\\s*"))
            outName = "mutModel";
        else
            outName = getID();
        for (int i=0; i<rateVector.getDimension(); i++) {
            out.print(outName + ".mutRateRaw_" + i + "\t");
        }
        for (int i=0; i<4; i++) {
            out.print(outName + ".freq_" + i + "\t");
        }
    }

    @Override
    public void log(long nSample, PrintStream out) {
    	double[] values= rateVector.getDoubleValues();
        double[] freqs= getNucleotideFrequencies();
    	for (int i=0; i<values.length ; i++) {
            out.print(values[i] + "\t");
        }
    	for (int i=0; i<4 ; i++) {
            out.print(freqs[i] + "\t");
        }
    }


	@Override
	public void close(PrintStream out) {
	}
	
    @Override
    protected boolean requiresRecalculation() {
        return false;
    }

    @Override
    protected void restore() {
        super.restore();
    }
    


}
