/*
* File TreeLikelihood.java
*
* Copyright (C) 2010 Remco Bouckaert remco@cs.auckland.ac.nz
*
* This file is part of BEAST2.
* See the NOTICE file distributed with this work for additional
* information regarding copyright ownership and licensing.
*
* BEAST is free software; you can redistribute it and/or modify
* it under the terms of the GNU Lesser General Public License as
* published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
*  BEAST is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with BEAST; if not, write to the
* Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
* Boston, MA  02110-1301  USA
*/

/* GRJ, Apr/May 2020: I edited code from the BADTRIP package (was TreeLikelihoodPoMoTransmission)
 */

package pomocov;

import java.lang.reflect.Constructor;
import java.util.*;

import beast.core.Description;
import beast.core.Distribution;
import beast.core.Input;
import beast.core.State;
import beast.core.Input.Validate;
import beast.core.parameter.RealParameter;
import beast.core.util.Log;

import pomocov.SiteModelInterfacePoMo.Base;
import beast.evolution.sitemodel.SiteModelInterface;
import beast.evolution.substitutionmodel.DefaultEigenSystem;
import beast.evolution.substitutionmodel.EigenDecomposition;
import beast.evolution.substitutionmodel.EigenSystem;

import beast.util.Randomizer;
import pomocov.utils.*;


@Description("Calculates the log-likelihood of deep sequencing data on a PoMo transmission tree given a site model and PoMo substitution model.")
public class PoMoCoVTreeLikelihood extends Distribution {
    
    final public Input<AlignmentPoMo> dataInput = new Input<>("data",
            "sequence PoMo data",
            Validate.REQUIRED);

    final public Input<TreePoMo> treePoMoInput = new Input<>("treePoMo",
            "phylogenetic transmission PoMo tree with sequence data in the leafs and internal nodes",
            Validate.REQUIRED);

    final public Input<SiteModelInterface> siteModelInput = new Input<>("siteModel",
            "site model for leafs in the beast.tree",
            Validate.REQUIRED);

    // GRJ TODO I doubt that it is useful to use anything except DefaultEigenSystem
    public Input<String> eigenSystemClass = new Input<>("eigenSystem",
            "Name of the class used for creating an EigenSystem",
            DefaultEigenSystem.class.getName());

    public Input<Boolean> useAmbiguitiesInput = new Input<>("useAmbiguities",
            "flag to indicate leafs that sites containing ambigue states should be handled instead of ignored (the default)",
            false);

    public Input<RealParameter> seqErrorInput = new Input<>("seqError",
            " starting sequencing error rate, 0 by default",
            Validate.REQUIRED);
    
	public Input<RealParameter> bottleneckInput = new Input<>("bottleneck",
            "parameter describing how many Ne generiations the bottleneck corresponds to in terms of drift",
            Validate.REQUIRED);

    enum Scaling {none, always, _default}

    public Input<Scaling> scaling = new Input<>("scaling",
            "type of scaling to use, one of " + Arrays.toString(Scaling.values()) + ". If not specified, the -beagle_scaling flag is used.",
            Scaling._default, Scaling.values());


    // debugging
	private final boolean debugFlag = Boolean.parseBoolean(System.getProperty("pomocov.debug"));
	// for debugging use           –enableassertions -Dpomocov.debug="true"
    private boolean debugPrintPartials = false;

    private TreePoMo tree;
    private AlignmentPoMo alignment;
    private PoMoGeneral substitutionModel;
    private SiteModelPoMo.Base siteModel;
    // flag to indicate ascertainment correction should be applied
    private boolean useAscertainedSitePatterns = false;
    // dealing with proportion of site being invariant
    private double proportionInvariant = 0;
    private List<Integer> constantPattern = null;

    private int nTaxa;
    private int nPatterns;
    private int virtualPopSize;
    private int nStates;

    // Parameters. (Others are in tree and site model.)
    private double sequencingError, sequencingErrorStored;
    private double bottleneck, bottleneckStored;

    // calculation engine
    private BeerLikelihoodCorePoMo likelihoodCore;

	private EigenDecomposition eigenDecompositionDrift2nucs;
	private EigenSystem eigenSystemDrift2nucs;
    // transition probabilities for part of bottleneck, for any two nucelotides. Used to
    // achieve exactly the same as probabilitiesBottleneck, but more efficiently.
    private double[] probabilitiesDrift2nucs, probabilitiesDrift2nucsStored;

    // For samples. These change when sequencingError changes
    private double[][] partialsTaxa, partialsTaxaStored;

    // For collecting samples, children, and infection events in a host
    private static class EventInHost {
        final int chIdx;
        final int smpIdx;
        final double height;
        double tipwardInterval;

        public EventInHost(int chIdx, int smpIdx, double height) {
            assert height >= 0.0;
            this.chIdx = chIdx;
            this.smpIdx = smpIdx;
            this.height = height;
            this.tipwardInterval = 0.0;
        }
    }

    // For calculating sample partials
    private class BiggestTwoCounts {
        final int count1;
        final int count2;
        final int nuc1;
        final int nuc2;

        public BiggestTwoCounts(int count1, int count2, int nuc1, int nuc2) {
            this.count1 = count1;
            this.count2 = count2;
            this.nuc1 = nuc1;
            this.nuc2 = nuc2;
        }
    }



    @Override
    public void initAndValidate() throws IllegalArgumentException {

        if (!((siteModelInput.get()) instanceof SiteModelPoMo.Base)) {
        	throw new IllegalArgumentException("siteModel input should be of type SiteModel.Base");
        }

        tree = treePoMoInput.get();
        alignment = dataInput.get();
        siteModel = (Base) siteModelInput.get();
        siteModel.setDataType((DataTypePoMo)(dataInput.get().getDataType()));
        substitutionModel = siteModel.substModelInput.get();
        nTaxa = alignment.getTaxonCount();
        nPatterns = dataInput.get().getPatternCount();
        virtualPopSize = substitutionModel.getPopSize();
        nStates = substitutionModel.getStateCount();
        assert nStates == IndicesEtc.nPoMoStates(virtualPopSize);

        proportionInvariant = siteModel.getProportionInvariant();
        siteModel.setPropInvariantIsCategory(false);
        if (proportionInvariant > 0) {
            calcConstantPatternIndices(nPatterns, nStates);
        }
        assert siteModel.getCategoryCount() <= 1; // GRJ pomocov does not support multiple gamma categories.

        sequencingError = seqErrorInput.get().getValue();
        sequencingErrorStored = sequencingError;
        partialsTaxa = new double[nTaxa][nPatterns * nStates];
        partialsTaxaStored = new double[nTaxa][nPatterns * nStates];

        if (dataInput.get().isAscertained) {
            useAscertainedSitePatterns = true;
        }

        likelihoodCore = new BeerLikelihoodCorePoMo(nStates);

        Log.info.println(getClass().getSimpleName() + "(" + getID() + ") uses " + likelihoodCore.getClass().getSimpleName());
        Log.info.println("  " + alignment.toString(true));        bottleneck = bottleneckInput.get().getValue();

        bottleneckStored = bottleneck;


        if (debugFlag) {
            BetaBinomial.test();
            TransmissionDistributions.sanityChecks();
        }

        double[][] rateMatrixDrift2nucs;
        rateMatrixDrift2nucs = getDrift2NucsRates(virtualPopSize);
        if (debugFlag) {
            DebugTune.printSigned2DMatrix(tree.debugGetStream(), "drift2nucs", rateMatrixDrift2nucs);
        }
		try {
			eigenSystemDrift2nucs = createDrift2nucsEigenSystem();
		} catch (Exception e) {
			e.printStackTrace();
		}
		eigenDecompositionDrift2nucs = eigenSystemDrift2nucs.decomposeMatrix(rateMatrixDrift2nucs);
		probabilitiesDrift2nucs = new double[(virtualPopSize + 1) *(virtualPopSize + 1)];
		Arrays.fill(probabilitiesDrift2nucs, 1.0);
		getTransitionProbabilitiesDrift2nucs();
		probabilitiesDrift2nucsStored = probabilitiesDrift2nucs.clone();
        if (debugFlag) {
            DebugTune.printBriefNonNeg1DMatrix(tree.debugGetStream(), "probabilitiesDrift2nucs", probabilitiesDrift2nucs);
        }

        List<String> tNames = alignment.getTaxaNames();
        for (int t = 0; t < nTaxa; t++){
            double[] currPartials = calcSamplePartials(alignment.getTaxonIndex(tNames.get(t)));
        	System.arraycopy(currPartials, 0, partialsTaxa[t], 0, currPartials.length);
    		System.arraycopy(currPartials, 0, partialsTaxaStored[t], 0, currPartials.length);
        }
        initCore();
    }





    /**
     * Calculate the log likelihood of the current state.
     *
     * @return the log likelihood.
     */
    @Override
    public double calculateLogP() {
        updateBottleneckTransitionProbabilities();
        updateSamplePartials();
        debugPrintPartials = debugFlag  &&  Math.abs(tree.debugGetCheckTreeNCalls() - 100000) < 3;
        // main recursion
        double[] rootPartials = calcPartialsSubtree(tree.getRoot(), nPatterns, tree.getUnobservedHosts());
        debugCheckPartials(rootPartials, " (unweighted) root partials");

        // approx integration (weighted sum) over rate categories at root, including invariant category
        double[] weightedRootPartials = new double[nPatterns * nStates];
        integrateRootPartialsOverRateCategories(rootPartials, weightedRootPartials);

        // Root partials to patternLLhoods[] using root frequencies
        double[] frequencies;
        if (substitutionModel.mutM.mutationOnlyModel) {
            double mu = siteModel.getRateForCategory(0, null);
            frequencies = substitutionModel.getBBRootFrequencies(mu, bottleneck);
        } else {
            frequencies = substitutionModel.getFrequencies();
        }
        double [] patternLLhoods = new double[nPatterns];
        likelihoodCore.calculateLogLikelihoods(weightedRootPartials, frequencies, patternLLhoods);

        //  Weighted sum of patternLLhoods[].
        logP = sumLogPOverPatterns(patternLLhoods);
        if (debugPrintPartials) {
            DebugTune.printRootFrequencies(tree.debugGetStream(), tree.debugGetCheckTreeNCalls(), virtualPopSize, frequencies);
            int [] weights = new int[nPatterns];
            for (int i = 0; i < nPatterns; i++) {
                weights[i] = dataInput.get().getPatternWeight(i);
            }
            DebugTune.printPatternWeightsLLhoods(tree.debugGetStream(), weights, patternLLhoods);
            tree.debugGetStream().println(tree.asText());
            tree.debugGetStream().println("logP " + logP);
        }
        return logP;
    }


    /**
     * check state for changed variables and update temp results if necessary *
     */
    @Override
    protected boolean requiresRecalculation() {
        return true;
        //System.out.println("Start requiresRecalculation ");

        //NICOLA TODO : allow beagle
    }


    @Override
    public void store() {
        if (sequencingError != sequencingErrorStored){
            for (int t=0; t<nTaxa; t++){
                System.arraycopy(partialsTaxa[t], 0, partialsTaxaStored[t], 0, nPatterns* nStates);
            }
        }
        sequencingErrorStored = sequencingError;
        if (likelihoodCore != null) {
            likelihoodCore.store();
        }
        super.store();

        /*TODO-Drift2nucs
        if (bottleneck!=bottleneckStored){
        	probabilitiesBottleneckStored=probabilitiesBottleneck.clone();
            probabilitiesDrift2nucsStored = probabilitiesDrift2nucs.clone();
        	bottleneckStored=bottleneck;
        }*/

    }

    @Override
    public void restore() {
        if (sequencingError != sequencingErrorStored){
            for (int t=0; t<nTaxa; t++){
                System.arraycopy(partialsTaxaStored[t], 0, partialsTaxa[t], 0, nPatterns* nStates);
            }
        }
        sequencingError = sequencingErrorStored;
        //partialsDirty=false;
        if (likelihoodCore != null) {
            likelihoodCore.restore();
        }
        super.restore();

        /*TODO-Drift2nucs
        if (bottleneck!=bottleneckStored){
        	probabilitiesBottleneck = probabilitiesBottleneckStored.clone();
        	probabilitiesDrift2nucs = probabilitiesDrift2nucsStored.clone();
        	bottleneckStored=bottleneck;
        }*/
    }

    /**
     * @return a list of unique ids for the state nodes that form the argument
     */
    public List<String> getArguments() {
        return Collections.singletonList(dataInput.get().getID());
    }

    /**
     * @return a list of unique ids for the state nodes that make up the conditions
     */
    public List<String> getConditions() {
        return siteModel.getConditions();
    }


    /**
     * This method samples the sequences based on the tree and site model.
     */
    @Override
    public void sample(State state, Random random) {
        throw new UnsupportedOperationException("Can't sample a fixed alignment!");
    }




    // *****************************************************************************
    // ***********************      private     ************************************
    // *****************************************************************************

    // *****************************************************************************
    // *****************        Initialisation      ********************************
    // *****************************************************************************



    /**
     * create an EigenSystem of the class indicated by the eigenSystemClass input *
     */
    private EigenSystem createEigenSystem() throws Exception {
        Constructor<?>[] ctors = Class.forName(eigenSystemClass.get()).getDeclaredConstructors();
        Constructor<?> ctor = null;
        for (int i = 0; i < ctors.length; i++) {
            ctor = ctors[i];
            if (ctor.getGenericParameterTypes().length == 1)
                break;
        }
        ctor.setAccessible(true);
        return (EigenSystem) ctor.newInstance(nStates);
    }


    // GRJ TODO. I think only ever want DefaultEigenSystem, no need to scan ctors
    private EigenSystem createDrift2nucsEigenSystem() throws Exception {
        Constructor<?>[] ctors = Class.forName(eigenSystemClass.get()).getDeclaredConstructors();
        Constructor<?> ctor = null;
        for (int i = 0; i < ctors.length; i++) {
            ctor = ctors[i];
            if (ctor.getGenericParameterTypes().length == 1)
                break;
        }
        ctor.setAccessible(true);
        return (EigenSystem) ctor.newInstance(virtualPopSize + 1);
    }


    /**
     * Determine indices of weightedRootPartials[] that need to be updates
     * // due to sites being invariant. If none of the sites are invariant,
     * // the 'site invariant' category does not contribute anything to the
     * // root probability. If the site IS invariant for a certain character,
     * // taking ambiguities in account, there is a contribution of 1 from
     * // the 'site invariant' category.
     */
    private void calcConstantPatternIndices(final int nPatterns, final int nStateCount) {
        constantPattern = new ArrayList<>();
        for (int i = 0; i < nPatterns; i++) {
           // final int[][] pattern = dataInput.get().getPatternPoMo(i);
            final boolean[] bIsInvariant = new boolean[nStateCount];
            Arrays.fill(bIsInvariant, true);
            for (int k = 0; k < nStateCount; k++) {
                if (bIsInvariant[k]) {
                    constantPattern.add(i * nStateCount + k);
                }
            }
        }
    }

    private void initCore() {
    	//TODO: move to later (calculateLogP) to account for new nodes added to the tree
        final int nodeCount = tree.getNodeCount();
        likelihoodCore.initialize(
                nodeCount,
                nPatterns,
                siteModel.getCategoryCount(),
                true, useAmbiguitiesInput.get()
        );
    }


    // *****************************************************************************
    // ******************** Main log lhood calculation *****************************
    // *****************************************************************************

    private void updateBottleneckTransitionProbabilities() {
        // Update bottleneck matrix if needed
        bottleneck = bottleneckInput.get().getValue();
        if (bottleneck != bottleneckStored){
            getTransitionProbabilitiesDrift2nucs();
        }
    }


    private void updateSamplePartials() {
        // update sample partials if needed
        sequencingError = seqErrorInput.get().getValue();
        if (sequencingError != sequencingErrorStored){
            List<String> tNames = alignment.getTaxaNames();
            for (int t=0; t<nTaxa; t++){
                double[] currPartials = calcSamplePartials(alignment.getTaxonIndex(tNames.get(t)));
                System.arraycopy(currPartials, 0, partialsTaxa[t], 0, nStates *nPatterns);
            }
        }
    }


    /**
     * Recursively calculate partials below node
     */
    private double[] calcPartialsSubtree(NodePoMo node, int patternCount, boolean unobservedHosts) {
        ArrayList<double[]> childrensPartials = new ArrayList<>();
        for (NodePoMo ch : node.getChildren()){
            double[] chPartials = calcPartialsSubtree(ch, patternCount, unobservedHosts);
            childrensPartials.add(chPartials);
        }
        double[] partials = calcNodePartials(node, childrensPartials, patternCount, unobservedHosts);
        return partials;
    }



    private ArrayList<EventInHost> collectEventsInHost(NodePoMo node, int nSamples, int nChildren) {
        ArrayList<EventInHost> EIHs = new ArrayList<>();
        int nIntervals = nChildren + nSamples;
        assert nIntervals > 0;
        for (int ch = 0; ch < nChildren; ch++) {
            EventInHost EIH = new EventInHost(ch, -1, node.getChild(ch).getHeight());
            EIHs.add(EIH);
        }
        for (int smp = 0; smp < nSamples; smp++) {
            EventInHost EIH = new EventInHost(-1, smp, node.getSample(smp).time);
            EIHs.add(EIH);
        }
        EventInHost EIH = new EventInHost(-1, -1, node.getHeight());
        EIHs.add(EIH);
        Collections.sort(EIHs, Comparator.comparing(item -> item.height));
        for (int eih = 1; eih < EIHs.size(); eih++) {
            EIHs.get(eih).tipwardInterval = EIHs.get(eih).height - EIHs.get(eih-1).height;
        }
        return  EIHs;
    }


    private String debugEventInHostAsText(int index, NodePoMo node, EventInHost EIH, int eih) {
        String eventText = "CheckTreeNCalls " + index + ". Host " + node.getHostName() + ". Event " + eih;
        if (EIH.chIdx >= 0) {
            eventText += ": child " + node.getChild(EIH.chIdx).getHostName() + " height " + EIH.height + " tipwardInterval " + EIH.tipwardInterval;
        } else if (EIH.smpIdx >= 0) {
            eventText += ": sample " + EIH.smpIdx + " height " + EIH.height + " tipwardInterval " + EIH.tipwardInterval;
        } else {
            eventText += ": infection at node, height " + EIH.height + " tipwardInterval " + EIH.tipwardInterval;
        }
        return eventText;
    }


    private double[] calcNodePartials(NodePoMo node, ArrayList<double[]> childrensPartials, int patternCount, boolean unobservedHosts) {
        List<Sample> samples = node.getSamples();
    	int nSamples = samples.size();
        int nChildren = node.getChildren().size();
        int nIntervals = nChildren+nSamples;
        if (nIntervals == 0){
        	double[] partial1 = new double[likelihoodCore.partialsSize];
        	Arrays.fill(partial1, 1.0);
        	return partial1;
        }
        ArrayList<EventInHost> EIHs = collectEventsInHost(node, nSamples, nChildren);
        double[] partials = new double[likelihoodCore.partialsSize];
     	double[] newPartials = new double[likelihoodCore.partialsSize];
        EventInHost EIH = EIHs.get(0);
        assert EIH.tipwardInterval == 0.0;
        // Start with the most tipward child or sample, get partials[]
        if (EIH.chIdx >= 0) {
            System.arraycopy(childrensPartials.get(EIH.chIdx), 0, partials, 0, patternCount* nStates);
            debugCheckPartials(partials, "1. most recent event (child)");
            if (debugPrintPartials) {
                String eventText = debugEventInHostAsText(tree.debugGetCheckTreeNCalls(), node, EIH, 0);
                DebugTune.printNodePartials(tree.debugGetStream(), eventText, patternCount, virtualPopSize, partials);
            }
        } else {
            assert EIH.smpIdx >= 0;
            int taxonIdx = alignment.getTaxonIndex(samples.get(EIH.smpIdx).getTaxon().getID());
            System.arraycopy(partialsTaxa[taxonIdx], 0, partials, 0, patternCount* nStates);
            debugCheckPartials(partials, "2. most recent event (sample)");
            if (debugPrintPartials){
                String eventText = debugEventInHostAsText(tree.debugGetCheckTreeNCalls(), node, EIH, 0);
                ArrayList<int[]> counts = new ArrayList<>();
                for (int patt = 0; patt < nPatterns; patt++) {
                    counts.add(alignment.getPatternPoMo(taxonIdx, patt));
                }
                DebugTune.printSampleAndPartials(tree.debugGetStream(), eventText, patternCount, virtualPopSize, counts, partials);
            }
        }

        // Intervals between children and samples. For each interval
        // Have p = partials[] from previous step
        // Make M = transProbsMatrix[,].
        // Get q = chOrSmpPartials[] = child partials or sample partials
        // newPartials[] = Mp * q where Mp is matrix-vector, and * is elementwise
        // copy newPartials[] to partials[]
        final double branchRate = siteModel.getRateForCategory(0, null);
        for (int eih = 1; eih < EIHs.size() - 1; eih++) {
            EIH = EIHs.get(eih);
            double [] chOrSmpPartials;
            if (EIH.chIdx >= 0) {
                chOrSmpPartials = childrensPartials.get(EIH.chIdx);
            } else {
                assert EIH.smpIdx >= 0;
                int taxonIdx = alignment.getTaxonIndex(samples.get(EIH.smpIdx).getTaxon().getID());
                chOrSmpPartials =partialsTaxa[taxonIdx];
            }
            calcPartialsForOneInterval(partials, EIH.tipwardInterval, branchRate, chOrSmpPartials, newPartials);
            System.arraycopy( newPartials, 0, partials, 0, newPartials.length );
            debugCheckPartials(partials, "3. a " + ((EIH.chIdx >= 0) ? "child" : "sample"));
            if (debugPrintPartials){
                String eventText = debugEventInHostAsText(tree.debugGetCheckTreeNCalls(), node, EIH, eih);
                DebugTune.printNodePartials(tree.debugGetStream(), eventText, patternCount, virtualPopSize, partials);
            }
        }
        // last interval to node
        EIH = EIHs.get(EIHs.size() - 1);
        calcPartialsForOneInterval(partials, EIH.tipwardInterval, branchRate, null, newPartials);
        System.arraycopy( newPartials, 0, partials, 0, newPartials.length );
        debugCheckPartials(partials, "4. last interval to node");
        if (debugPrintPartials){
            String eventText = debugEventInHostAsText(tree.debugGetCheckTreeNCalls(), node, EIH, EIHs.size() - 1);
            DebugTune.printNodePartials(tree.debugGetStream(), eventText, patternCount, virtualPopSize, partials);
        }

        if (unobservedHosts  &&  !node.isRoot()) {
            double duration = node.getOriginHeight() - node.getHeight();
            calcUnobservedChainPartials(duration, node.getUnobservedCount(), branchRate, partials, newPartials);

            if (debugPrintPartials){
                double g = node.getOriginHeight();
                double h = node.getHeight();
                String eventText =  "CheckTreeNCalls " + tree.debugGetCheckTreeNCalls() + ". Host " + node.getHostName();
                eventText += ". Bottleneck of intensity " + bottleneck + " and " + node.getUnobservedCount() + " unob hosts ";
                eventText += "originHeight " + g + " height " + h + " interval " + (g-h);
                DebugTune.printNodePartials(tree.debugGetStream(), eventText, patternCount, virtualPopSize, newPartials);
            }
        } else {
            calcBottleneckPartials(partials, newPartials);
            if (debugPrintPartials){
                String eventText =  "CheckTreeNCalls " + tree.debugGetCheckTreeNCalls() + ". Host " + node.getHostName();
                eventText += ". Bottleneck (no unobserved hosts)";
                DebugTune.printNodePartials(tree.debugGetStream(), eventText, patternCount, virtualPopSize, newPartials);
            }
        }
        debugCheckPartials(newPartials, "5. unobserved chain");

        return newPartials;
    }



    private void calcUnobservedChainPartials(double duration, int uc, double branchRate,
                                             double [] partials, double [] newPartials) {
        double delayScale = pomocov.utils.TransmissionDistributions.medianDelayTime();
        double inHostScale = pomocov.utils.TransmissionDistributions.medianWithinHostTime();
        double delayWeight = (uc + 1.0) * delayScale;
        double inHostWeight = uc * inHostScale;
        double inHostDuration = duration * inHostScale / (delayWeight + inHostWeight);
        calcBottleneckPartials(partials, newPartials);
        for (int i =0; i < uc; i++) {
            System.arraycopy( newPartials, 0, partials, 0, newPartials.length );
            calcPartialsForOneInterval(partials, inHostDuration, branchRate, null, newPartials);
            System.arraycopy( newPartials, 0, partials, 0, newPartials.length );
            calcBottleneckPartials(partials, newPartials);
        }
    }


    // *****************************************************************************
    // ******************* Low level log lhood calculation *************************
    // *****************************************************************************

    // samples to partials
    // calc partials for one interval to sample, child, or node
    // calc partials for bottlenecks
    // approx integration (weighted sum) over rate categories at root
    // Multiply (sum logs of) partials at root over patterns

    private double[] calcSamplePartials(int iTaxon) {
        double [] partials = new double[nPatterns*nStates];
        boolean debugTie = false;//  ties are broken at random so check is tricky
        for (int patt = 0; patt < nPatterns; patt++) {
            int [] acgtCounts =  alignment.getPatternPoMo(iTaxon, patt);
            if (debugFlag) {
                debugTie = debugTie || ACGTCountsPartial.debugRandomisationRequired(acgtCounts);
            }
            ACGTCountsPartial.calcSamplePartialsForOnePattern(acgtCounts, virtualPopSize, sequencingError, partials, patt);
        }
        return partials;
    }



    private void  calcPartialsForOneInterval(double[] partials, double interval, double branchRate,
                                             double[] chOrSmpPartials, double[] newPartials) {
        double [] debugNewPartials = new double[newPartials.length];


        if (substitutionModel.mutM.mutationOnlyModel) {
            double[] mutRates = substitutionModel.mutM.getACtoTGratesNormalisedForLinearApprox();
            double [] mutPs = new double[12]; // Order: A->C, A->G, A->T, C->A, etc...
            for (int i = 0; i < 12; i++) {
                mutPs[i] = mutRates[i] * interval * branchRate;
                // prevent numerical problems in case MCMC visits strange values
                mutPs[i] = mutPs[i] / (1.0 + 3 * mutPs[i]);
            }
            if (chOrSmpPartials != null) {
                likelihoodCore.calcPartialsToChOrSmp_Linear(partials, mutPs, chOrSmpPartials, newPartials, virtualPopSize);
            }  else {
                likelihoodCore.calcPartialsOneBranch_Linear(partials, mutPs, newPartials, virtualPopSize);
            }
        } else {
            double[] transProbsMatrix = new double[likelihoodCore.matrixSize];
            substitutionModel.getTransitionProbabilities(interval, branchRate, transProbsMatrix);
            if (chOrSmpPartials != null) {
                likelihoodCore.calculatePartialsFixedPruning(partials, transProbsMatrix, chOrSmpPartials, newPartials);
            } else {
                likelihoodCore.calculatePartialsPruning(partials, transProbsMatrix, newPartials);
            }
            if (debugFlag  && tree.debugGetCheckTreeNCalls() < 10) {
                DebugTune.printNonNeg1DMatrix(tree.debugGetStream(),"calcPartialsForOneInterval() : transProbsMatrix", transProbsMatrix);
            }
        }
    }




    private void calcBottleneckPartials(double [] partials, double [] newPartials) {
        if (virtualPopSize == 3) {
            // experimental likelihoodCore.calcPartialsBottle_Binomial_V3(partials, Bottleneck.bottleneckForVPop3(bottleneck), newPartials);
            likelihoodCore.calcPartialsBottle_D2N_V3(partials, probabilitiesDrift2nucs, newPartials);
        } else {
            likelihoodCore.calcPartialsBottle_D2N(partials, probabilitiesDrift2nucs, newPartials, virtualPopSize);
        }
    }


    private void integrateRootPartialsOverRateCategories(double[] rootPartials, double[] weightedRootPartials) {
        // Deal with multiple rate categories. Currently (June 2020) does nothing
        // weightedRootPartials[] are weighted per rate category, but this is not yet implemented.
        // So calculateIntegratePartials() just copies.
        // Also I never use constantPattern
        double[] proportions = siteModel.getCategoryProportions();
        assert proportions.length == 1;
        assert proportions[0] == 1.0;
        likelihoodCore.calculateIntegratePartials(rootPartials, proportions, weightedRootPartials);
        debugCheckPartials(weightedRootPartials, "weighted root partials");
        if (constantPattern != null) {
            proportionInvariant = siteModel.getProportionInvariant();
            for (final int i : constantPattern) {
                weightedRootPartials[i] += proportionInvariant;
            }
        }
    }


    private double sumLogPOverPatterns(double [] patternLLhoods) {
        double sumLogP = 0.0;
        if (useAscertainedSitePatterns) {
            final double ascertainmentCorrection = dataInput.get().getAscertainmentCorrection(patternLLhoods);
            assert Double.isFinite(ascertainmentCorrection);
            for (int i = 0; i < nPatterns; i++) {
                sumLogP += (patternLLhoods[i] - ascertainmentCorrection) * dataInput.get().getPatternWeight(i);
                assert Double.isFinite(sumLogP);
            }
        } else {
            for (int i = 0; i < nPatterns; i++) {
                sumLogP += patternLLhoods[i] * dataInput.get().getPatternWeight(i);
            }
        }
        assert Double.isFinite(sumLogP);
        if (sumLogP>0.0){
        	System.out.println("log-likelihood>0, probably matrix exponentiation problem, rejecting move\n");
            sumLogP=Double.NEGATIVE_INFINITY;
        }
        return sumLogP;
    }




    // *****************************************************************************
    // ************************ Making bottleneck matrices *************************
    // *****************************************************************************


    // Converts real bottleneck b into a bottleneck intensity B as used by BADTRIP.
    // A bottleneck of size b means random sampling of b virions (giving a transition
    // matrix P(b) containing binomial distributions). Suppose first that b = v, where
    // v is the size of the virtual population. This bottleneck is similar to, but not
    // the same as, exponentiating the Moran-style drift rate matrix R (with entries
    // like i(v-i)/v^2), with B = v/2. That is, P(v) ~= exp((v/2)R).
    // (I checked the formula experimentally, I believe B ~= v/2 can be proved.)
    // Then B is scaled to account for the difference between b and v. The scaling
    // is based on the observation that
    // Pr(not coalesced after n generations in population size N) = ((N-1)/N)^n
    private double bottleneckIntensity() {
        assert bottleneck >= 1.0;
        double b = Math.max(bottleneck, 1.01);
        double v = virtualPopSize;
        double B = 0.5 * v;
        B *= Math.log(b/(b-1.0));
        B /= Math.log(v/(v-1.0));
        return B;
    }


	private void getTransitionProbabilitiesDrift2nucs() {
    	int dim = virtualPopSize + 1;
		double[] iexp = new double[dim * dim];
		// Eigen vectors
		double[] Evec = eigenDecompositionDrift2nucs.getEigenVectors();
		// inverse Eigen vectors
		double[] Ievc = eigenDecompositionDrift2nucs.getInverseEigenVectors();
		// Eigen values
		double[] Eval = eigenDecompositionDrift2nucs.getEigenValues();
		double B = bottleneckIntensity();
		for (int i = 0; i < dim; i++) {
			double temp = Math.exp(B * Eval[i]);
			for (int j = 0; j < dim; j++) {
				iexp[i * dim + j] = Ievc[i * dim + j] * temp;
			}
		}
		int u = 0;
		for (int i = 0; i < dim; i++) {
			for (int j = 0; j < dim; j++) {
				double temp = 0.0;
				for (int k = 0; k < dim; k++) {
					temp += Evec[i * dim + k] * iexp[k * dim + j];
				}
                probabilitiesDrift2nucs[u] = Math.abs(temp);
				u++;
			}
		}
		//if exponentiation fails, die
		for (int i = 0; i < dim; i++) {
			double temp = 0.0;
			for (int j = 0; j < dim; j++) {
				temp += probabilitiesDrift2nucs[i*dim +j];
			}
			assert temp < 1.001 && temp > 0.999;
		}
	}



	// Part of the drift-only rate matrix
	private double[][] getDrift2NucsRates(int vPop){
		int dim = vPop + 1;
		double[][] rates = new double[dim][dim];
		for(int i = 1; i < vPop; i++){
			double rate =  (double)i * (double)(vPop - i);
			rate /= (vPop * vPop);
			rates[i][i-1] = rates[i][i+1] = rate;
			rates[i][i] = - 2.0 * rate;
		}
		return rates;
	}



    // *****************************************************************************
    // ************************ Debugging and tuning *******************************
    // *****************************************************************************




    private void debugCheckPartialsMatch(double[] ps1, double[] ps2, String msg) {
        for (int i = 0; i < ps1.length; i++) {
            double sum = ps1[i] + ps2[i];
            double diff = ps1[i] - ps2[i];
            if (Math.abs(diff) > 1e-9) {
                System.err.println("Partials mismatch. " + msg);
                System.exit(1);
            }
        }
    }



    private void debugCheckPartials(double[] partial, String message) {
        if (debugFlag) {
            double sum=0.0;
            for (int i = 0; i < partial.length; i++) {
                if (!Double.isFinite(partial[i])) {
                    System.err.println("NAN in checkPartials() i = " + i + " " + message);
                    System.exit(1);
                }
                sum += partial[i];
            }
            if (sum <= 0.0) {
                System.err.println("Warning checkPartials() found partials all zero (underflow?)" + message);
            }
            assert sum > 0.0;
        }
    }



} // class TreeLikelihood
