package pomocov.utils;

import java.util.Arrays;

public class RootFrequencies {



    // GRJ This uses a different method from BADTRIP to calculate the root frequncies.
    //
    // The theta value (2 N_e mu) is estimated from the bottleneck (which is N_e) and the mutation
    // rate (per site per day) multiplied by the typical time within a host. This gives the
    // mutations per generation (which is mu).

    // The Moran drift and mutation model for two alleles has a beta-binomial equilibrium distribution which
    // can be calculated exactly. The BetaBinomial class does that.
    // For the PoMoCoV model, with drift-only followed by mutation-only, and with root not the equilibrium,
    // I approximate as follows.
    // I assume the frequencies for the 4 nucleotides are given (perhaps empirical later).
    // For each of the 6 edges, 1/3 of the 2 nucleotide frequencies is assigned to the edge (the 2 vertices
    // of the edge and the (vPop-1) polymorphic states along the edge). Then a beta-binomial is fitted.
    public static void fillBBRootFrequencies(double[] nucFreqs, double[] ACtoTGrates, double mu, double bottleneck, int vPop, double[] bbrfs) {
        double theta = 2.0 * bottleneck * mu * TransmissionDistributions.medianWithinHostTime();
        Arrays.fill(bbrfs, 0.0);
        fillBetweenTwoNucs(bbrfs, IndicesEtc.EdgeIndicesForAC(vPop), ACtoTGrates[IndicesEtc.rAC], ACtoTGrates[IndicesEtc.rCA], theta, vPop, nucFreqs);
        fillBetweenTwoNucs(bbrfs, IndicesEtc.EdgeIndicesForAG(vPop), ACtoTGrates[IndicesEtc.rAG], ACtoTGrates[IndicesEtc.rGA], theta, vPop, nucFreqs);
        fillBetweenTwoNucs(bbrfs, IndicesEtc.EdgeIndicesForAT(vPop), ACtoTGrates[IndicesEtc.rAT], ACtoTGrates[IndicesEtc.rTA], theta, vPop, nucFreqs);
        fillBetweenTwoNucs(bbrfs, IndicesEtc.EdgeIndicesForCG(vPop), ACtoTGrates[IndicesEtc.rCG], ACtoTGrates[IndicesEtc.rGC], theta, vPop, nucFreqs);
        fillBetweenTwoNucs(bbrfs, IndicesEtc.EdgeIndicesForCT(vPop), ACtoTGrates[IndicesEtc.rCT], ACtoTGrates[IndicesEtc.rTC], theta, vPop, nucFreqs);
        fillBetweenTwoNucs(bbrfs, IndicesEtc.EdgeIndicesForGT(vPop), ACtoTGrates[IndicesEtc.rGT], ACtoTGrates[IndicesEtc.rTG], theta, vPop, nucFreqs);
    }


    private static void fillBetweenTwoNucs(double[] bbrfs, IndicesEtc.EdgeIndices eis, double rXY, double rYX, double theta, int vPop, double[] nucFreqs) {
        // do a third of each nuc at each call
        double pXY = (nucFreqs[eis.nucX] + nucFreqs[eis.nucY]) / 3.0;
        double [] probs = BetaBinomial.probabilities(vPop, theta*rXY, theta*rYX);
        bbrfs[eis.nucX] += pXY * probs[0];
        for (int i = 0; i < vPop - 1; i++) {
            bbrfs[eis.sX1Yn + i] = pXY * probs[i+1];
        }
        bbrfs[eis.nucY] += pXY * probs[vPop];
    }



}
